#include <string.h>
#include <malloc.h>
#include <iostream.h>

#include "balancer.h"

/* -------------  HASH TABLE MANAGEMENT FUNCTIONS  ------------------------------- */

/*
 the balancer hash table
*/
struct bal_task_entry *bal_table[HASH_TABLE_SIZE];


/*
 the hashing function, returns the position of the string str in the hashing table,
 reported by dan bernstein in comp.lang.c
*/
unsigned long bal_hash(unsigned char *str){
  unsigned long hash = 5381;
  int c;
  while (c = *str++)
    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

  return hash % HASH_TABLE_SIZE;
}

/*
 looks for name in the hash table, if it finds one, returns the pointer to it,
 otherwise NULL
*/
struct bal_task_entry *bal_search(char *name){
  unsigned long hash=bal_hash((unsigned char *)name);
  struct bal_task_entry *tmp;

  tmp=bal_table[hash];

  while(tmp!=NULL){
    if(!strncmp(tmp->task_name,name,TASK_NAME_LENGTH))return tmp;
    tmp=tmp->next;
  }
  return NULL;
}

/*
 creates a new entry in the hash table for the given name and returns
 the pointer to it. assumes that the entry in the table does not exist ...
*/
struct bal_task_entry *bal_insert(char *name){
  unsigned long hash=bal_hash((unsigned char *)name);
  struct bal_task_entry *tmp=(struct bal_task_entry*)malloc(sizeof(struct bal_task_entry));
  struct bal_task_entry **store;

  if(tmp==NULL)return NULL;
  
  store=&(bal_table[hash]);
  while(*store!=NULL){
    store=&((*store)->next);
  }
  *store=tmp;
  
  strncpy(tmp->task_name,name,TASK_NAME_LENGTH);
  tmp->lastupdate=time(NULL);
  tmp->cpu_pct_avg=0;
  tmp->total_avg=0;
  tmp->mem_avg=0;
  tmp->of_writes=0;

  return tmp;
}

/*
 deletes an entry from the hash table ... it first finds it and then deletes 
*/
void bal_delete(char *name){
  unsigned long hash=bal_hash((unsigned char *)name);
  struct bal_task_entry *tmp,*last=NULL;

  tmp=bal_table[hash];

  while(tmp!=NULL){
    if(!strncmp(tmp->task_name,name,TASK_NAME_LENGTH)){
	if(last!=NULL)last->next=tmp->next;
	  else bal_table[hash]=tmp->next;
	free(tmp);
	return;
    }
    last=tmp;
    tmp=tmp->next;
  }
}

/*
 initializes the hash table, must be called at startup
*/
void bal_table_init(){
  int i;
  for(i=0;i<HASH_TABLE_SIZE;i++)bal_table[i]=NULL;
}

/* ----------------- DATA UPDATE FUNCTIONS ------------------------ */

/*
 updates the data in the hash table, possibly creating the entry if it does not exist yet
 this one is called at the termination of a task
*/
void bal_update_data(char *name,double pct_cpu,double lifespan,double mem_res){
  struct bal_task_entry *tmp;

  if((tmp=bal_search(name))==NULL){
    tmp=bal_insert(name);
  }
  
  cerr<<endl<<"BALANCER UPDATE pname="<<name<<" pct_cpu="<<pct_cpu<<" lifespan="<<lifespan<<" mem_res="<<mem_res<<" writes="<<tmp->of_writes<<endl;
  
  tmp->cpu_pct_avg=(tmp->cpu_pct_avg*tmp->of_writes+pct_cpu)/(tmp->of_writes+1);

  tmp->total_avg=(tmp->total_avg*tmp->of_writes+lifespan)/(tmp->of_writes+1);

  tmp->mem_avg=(tmp->mem_avg*tmp->of_writes+mem_res)/(tmp->of_writes+1);
  
  tmp->lastupdate=time(NULL);
  
  tmp->of_writes++;
}


/*
 this function cleans up old entries in the table 
*/
void bal_cleanup_data(){
  int i;
  struct bal_task_entry *tmp,*last;
  time_t t=time(NULL)-3600*ENTRY_EXPIRE_TIME;
  
  for(i=0;i<HASH_TABLE_SIZE;i++){
    tmp=bal_table[i];
    last=NULL;
    while(tmp!=NULL){
      if(tmp->lastupdate<t){
	// now we delete this entry
	if(last!=NULL)last->next=tmp->next;
	  else bal_table[i]=tmp->next;
	free(tmp);
	if(last!=NULL)tmp=last->next;
	  else tmp=bal_table[i];  
      }else{
	last=tmp;
        tmp=last->next;
      }
    }
  }
}

struct bal_dead_queue_entry *bal_dead_queue=NULL;
struct bal_dead_queue_entry *bal_dead_queue_end=NULL;

/*
 insert a terminated task in the wait for dead queue ...
*/
void bal_queue_insert(char *name,int tid,double pct_cpu,double lifespan,double mem_res){
  struct bal_dead_queue_entry *tmp;
  
  cerr<<endl<<"WAIT FOR DEAD QUEUE INSERT pname="<<name<<" pct_cpu="<<pct_cpu<<" lifespan="<<lifespan<<" mem_res="<<mem_res<<" tid="<<tid<<endl;

  tmp=(struct bal_dead_queue_entry *)malloc(sizeof(struct bal_dead_queue_entry));
  
  strcpy(tmp->task_name,name);
  tmp->tid=tid;
  tmp->pct_cpu=pct_cpu;
  tmp->lifespan=lifespan;
  tmp->mem_res=mem_res;
  tmp->expire=time(NULL)+BAL_QUEUE_EXPIRE;
  tmp->prev=NULL;

  if(bal_dead_queue==NULL)bal_dead_queue=tmp;
  if(bal_dead_queue_end!=NULL)bal_dead_queue_end->prev=tmp;
  bal_dead_queue_end=tmp;
}

/*
 checks the wait for dead queue and if it finds some expired tasks, then it removes them and adds them into the HASH table
*/
void bal_queue_purge(){
  struct bal_dead_queue_entry *tmp,*last=NULL;
  
  tmp=bal_dead_queue;
  
  while(tmp!=NULL){
    /* did this one expire ??? */
    if(tmp->expire<time(NULL)){
      /* update the HASH table ... */
      bal_update_data(tmp->task_name,tmp->pct_cpu,tmp->lifespan,tmp->mem_res);

      /* remove this entry from the queue */
      if(last!=NULL)last->prev=tmp->prev;
      if(bal_dead_queue==tmp)bal_dead_queue=tmp->prev;
      if(bal_dead_queue_end==tmp)bal_dead_queue_end=last;
      free(tmp);
      if(last==NULL)last=bal_dead_queue;
    }else last=tmp;
    
    if(last!=NULL)tmp=last->prev; else tmp=NULL;
  }
}

/*
 checks the wait for dead queue if there is this tid and if it finds one, it deletes the entry automatically and
 adds the data to the process statistics
 returns 1 on found and 0 on not found
*/
int bal_queue_search(int tid,struct procs_data *data){
  struct bal_dead_queue_entry *tmp,*last=NULL;
  
  tmp=bal_dead_queue;
  
  while(tmp!=NULL){
    /* is this the right tid ??? */
    if(tmp->tid==tid){
      /* remove this entry from the queue */
      if(last!=NULL)last->prev=tmp->prev;
      if(bal_dead_queue==tmp)bal_dead_queue=tmp->prev;
      if(bal_dead_queue_end==tmp)bal_dead_queue_end=last;

  cerr<<endl<<"RESURRECTED task pname="<<tmp->task_name<<" pct_cpu="<<tmp->pct_cpu<<" lifespan="<<tmp->lifespan<<" mem_res="<<tmp->mem_res<<" tid="<<tmp->tid<<endl;
      
      data->total_pct_cpu=
                      ((data->total_pct_cpu*data->totals)+
                       (tmp->pct_cpu*tmp->lifespan))/
                       (data->totals+tmp->lifespan);
      data->totals+=tmp->lifespan;
      
      free(tmp);
      return 1;
    }else last=tmp;
    
    tmp=last->prev;
  }
  /* hmm, not found :(*/
  return 0;
}

