// #define DEBUG 1

//#include <iostream.h>
#include <fstream.h>

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#ifndef SIMULATOR
#include <unistd.h>
#endif

#include <time.h>
#include <signal.h>

#include <stddef.h>
#include <string.h>
#include <vector.h>
#include <algorithm>

#ifndef _RWSTD_NO_NAMESPACE
using namespace std;
#endif

#ifndef SIMULATOR
#include <pthread.h>
#include "pvm3.h"

#include "monitor.h"
#endif // SIMULATOR

#include "migrator.h"

//== Weight factors for load dimensions ==
double CpuWeight = 1.0;
double MemWeight = 1.0;
double MigWeight = 0.3;

//== Relative performance factors for the various hosts
vector<double> MemFactor;
vector<double> CpuFactor;

//== Flag that indicates if there is still time available for the search
short TimeExpired;


#ifndef SIMULATOR
/*=========================================================================*/
/* Input: hosts       -  number of available hosts
 * Output: Cpu - array of CPU weight factors as set in the config file
 *         Mem - array of memory weight factors as set in the config file
 */
void
SetCpuAndMemFactors( vector<double>& Cpu, vector<double>& Mem, unsigned hosts )
{
    extern resdes* rd_base;
    extern t_global_options global_options;
    extern char* configfile;

    read_options( &global_options, configfile );

    MemFactor = vector<double>( hosts, 1.0 );
    CpuFactor = vector<double>( hosts, 1.0 );

    /* Determine host id of this host from the resource entries */
    for (resdes* tmp_rd = rd_base; tmp_rd != NULL; tmp_rd = tmp_rd->next) {
        assert( tmp_rd->hid > 0 );
        assert( tmp_rd->hid <= hosts );
        CpuFactor[ tmp_rd->hid-1 ] = tmp_rd->cpu_weight;
        MemFactor[ tmp_rd->hid-1 ] = tmp_rd->mem_weight;
    }

   CpuWeight = global_options.CPUweight;
   MemWeight = global_options.MEMweight;
   MigWeight = global_options.MIGweight;
}





/*=========================================================================*/
/* Find the best mapping
 * Input: host_stats  -  statistics about all  PVM tasks
 *        lm_stats    -  statistics about all available nodes
 *        hosts       -  number of available hosts
 * Output: mig_vec    - vector of all necessary migrations
 * Returns: Number of necessary migrations
 *
 */
int
FindBestMapping( procs_hosts_stats* task_stats, load_stats** lm_stats, unsigned hosts,
                 vector<TMigration>& mig_vec )
{
    extern t_global_options global_options;
    // initialize performance factors
    SetCpuAndMemFactors( CpuFactor, MemFactor, hosts );

    // Stop time needed for computation
    time_t Started = time(0);


    unsigned MigrationsNecessary = 0;
    mig_vec = vector<TMigration>();

    // Collect information about all moveable PVM-tasks
    vector<TTaskInfo> TaskInfo;
    GetTaskInfo( task_stats, TaskInfo, hosts );
    if( TaskInfo.size() == 0 )  {
        cerr << "No migratable tasks found (1)" << endl;
        return 0;
    }

    // Load information on all available hosts
    vector<THostInfo> HostInfo( hosts );
    GetHostInfo( lm_stats, hosts, HostInfo, TaskInfo );

    EliminateDeadTasks( TaskInfo, HostInfo );

    HostInfoSubstractPVMLoad( HostInfo, TaskInfo );

    if( TaskInfo.size() == 0 )  {
        cerr << "No migratable tasks found (2)" << endl;
        return 0;
    }

    cerr << "Migration cost is " << dec<<MigWeight << endl
         << "CPU weight factor is " << CpuWeight << endl
         << "Memory weight factor is " << MemWeight << endl
         << "Maximum allowed search time is " << dec<<global_options.MaxSearchTime << " seconds" << endl;
    cerr << TaskInfo.size() << " tasks found" << endl;
    for( int i=0; i<TaskInfo.size(); i++ ) {
      cerr << i << ") Host: " << TaskInfo[i].Host << " TID: " << hex<<TaskInfo[i].PVMtid
             << " CPU: " << TaskInfo[i].CPUload << " MEM: " << TaskInfo[i].MEMload
             << endl;
    }

    // Holds the partial solution during the branch and bound run
    vector<unsigned> CurrentMapping( TaskInfo.size(), 0 );

    // Best mapping found so far
    vector<unsigned> BestMapping;

    // Costs of the best valid solution that has so far been found
    double Bound = -1;

    /* Call branch&bound routine */
    vector<double> PVM_CPUload( HostInfo.size(), 0 );
    vector<double> PVM_MEMload( HostInfo.size(), 0 );
    unsigned Nodes = 0;
    TimeExpired = FALSE;
    BB( TaskInfo, HostInfo, Bound, 0, CurrentMapping, 0, 0, 0, PVM_CPUload, PVM_MEMload,
        BestMapping, Nodes );
    // Cancel the alarm timer
    alarm( 0 );
    if( TimeExpired )   cerr << "Maximum search time expired " << endl;
    cerr << "Number of nodes checked: " << dec<<Nodes << endl;

    /* Determine the necessary migrations */
    MigrationsNecessary = DetermineMigrations( TaskInfo, HostInfo, BestMapping, mig_vec );

    cerr << "Time needed for BB-search: " << time(0)-Started << " secs." << endl;
    return MigrationsNecessary;
}


/*
 * Score this particular migration
 * 
 * the higher the returned score, the better the migration
 */
double
ScoreMigration(const vector<TTaskInfo>& Tasks, const vector<THostInfo>& Hosts,int my_src,int my_dest,int my_task){
    double score;
    double load_src_new,load_dest_new;
        
#ifdef USE_HASH_DATA


    /* if this task would make the destination even more loaded than the source, then skip it right away ... */
    if(Hosts[my_dest].NonPVM_CPUload+Hosts[my_dest].PVM_CPUload+Tasks[my_task].CPUload+NODE_MINIMAL_IMPROVE+Hosts[my_dest].trend>=
       Hosts[my_src].NonPVM_CPUload+Hosts[my_src].PVM_CPUload)return -1;

    load_src_new=Hosts[my_src].NonPVM_CPUload+Hosts[my_src].PVM_CPUload-Tasks[my_task].CPUload + Hosts[my_src].trend;
    load_dest_new=Hosts[my_dest].NonPVM_CPUload+Hosts[my_dest].PVM_CPUload+Tasks[my_task].CPUload + Hosts[my_dest].trend;

    score=1-((load_src_new/NODE_OVERLOAD_TRESHOLD)+(load_dest_new/NODE_OVERLOAD_TRESHOLD))/2;

    score*=Tasks[my_task].CPUload;

#else
    /* if this task would make the destination even more loaded than the source, then skip it right away ... */
    if(Hosts[my_dest].NonPVM_CPUload+Hosts[my_dest].PVM_CPUload+Tasks[my_task].CPUload+NODE_MINIMAL_IMPROVE+Hosts[my_dest].trend>=
       Hosts[my_src].NonPVM_CPUload+Hosts[my_src].PVM_CPUload)return -1;

    load_src_new=Hosts[my_src].NonPVM_CPUload+Hosts[my_src].PVM_CPUload-Tasks[my_task].CPUload + Hosts[my_src].trend;
    load_dest_new=Hosts[my_dest].NonPVM_CPUload+Hosts[my_dest].PVM_CPUload+Tasks[my_task].CPUload + Hosts[my_dest].trend;

    score=1-((load_src_new/NODE_OVERLOAD_TRESHOLD)+(load_dest_new/NODE_OVERLOAD_TRESHOLD))/2;

    score*=Tasks[my_task].CPUload;
#endif
    return score;
}



int 
UpdateMigrationData(int src,int dest,int task,vector<TTaskInfo>& TaskInfo,
            vector<THostInfo>& HostInfo){

  /* we will remove the task from the old node and transfer it to the new node ...*/
  TaskInfo[task].Host=dest;

  
  HostInfo[src].PVM_CPUload-=TaskInfo[task].CPUload / CpuFactor[src];
  HostInfo[src].PVM_MEMload-=TaskInfo[task].MEMload / MemFactor[src];
  
  HostInfo[src].source=1;				  

  HostInfo[dest].PVM_CPUload += TaskInfo[task].CPUload / CpuFactor[dest];
  HostInfo[dest].PVM_MEMload += TaskInfo[task].MEMload / MemFactor[dest];

  HostInfo[dest].dest=1;

  return 1;
}
	    

/*=========================================================================*/
/* Find a migration 
 * Input: host_stats  -  statistics about all  PVM tasks
 *        lm_stats    -  statistics about all available nodes
 *        hosts       -  number of available hosts
 
 * Output: migration    -  the wanted migration
 * Returns: 0 if no migration, 1 if one found
 *
 */
int
FindMigrations( procs_hosts_stats* task_stats, load_stats** lm_stats, unsigned hosts,
                 vector<TMigration>& mig_vec  , int maxmig)
{
    extern t_global_options global_options;
    // initialize performance factors
    SetCpuAndMemFactors( CpuFactor, MemFactor, hosts );

    double max_load=-1;
    int found_mig=0;

    // Collect information about all moveable PVM-tasks
    vector<TTaskInfo> TaskInfo;
    GetTaskInfo( task_stats, TaskInfo, hosts );

    if(TaskInfo.size()==0){
        cerr << "No migratable tasks found (1)" << endl;
        return 0;
    }

    // Load information on all available hosts
    vector<THostInfo> HostInfo( hosts );
    GetHostInfo( lm_stats, hosts, HostInfo, TaskInfo );

    EliminateDeadTasks( TaskInfo, HostInfo );

    HostInfoSubstractPVMLoad( HostInfo, TaskInfo );

    if(TaskInfo.size()==0){
        cerr << "No migratable tasks found (2)" << endl;
        return 0;
    }

    cerr << "Migration cost is " << dec<<MigWeight << endl
         << "CPU weight factor is " << CpuWeight << endl
         << "Memory weight factor is " << MemWeight << endl
	 << "Node OVERLOAD_TRESHOLD value is "<< NODE_OVERLOAD_TRESHOLD << endl
         << "Maximum allowed search time is " << dec<<global_options.MaxSearchTime << " seconds" << endl;
    cerr << TaskInfo.size() << " tasks found" << endl;
    for( int i=0; i<TaskInfo.size(); i++ ) {
      cerr << i << ") Host: " << TaskInfo[i].Host << " Name: " << TaskInfo[i].task_name << " TID: " << hex<<TaskInfo[i].PVMtid
             << " CPU: " << TaskInfo[i].CPUload << " MEM: " << TaskInfo[i].MEMload << " AVG_CPU: " << TaskInfo[i].CPUload_avg
	     << " LIFESPAN: " << TaskInfo[i].lifespan
	     << endl;
    }


    /* find an eligible host with the biggest load [ total ]*/
    int i=0;
    int j=0;
    int k=0;
    
    /* the best node picked as source for migration */
    int best_src_node;
    int best_dest_node;
    int best_task;
    
    /* the load on the best node */
    double best_score;
    double score;
    
    int m=maxmig;

    while(m>0){
      best_score=0;
      best_src_node=-1;
      best_dest_node=-1;
      best_task=-1;

      for(i=0;i<HostInfo.size();i++){
        if(HostInfo[i].Name!=""){
  	  if(!HostInfo[i].dest){
	    if((HostInfo[i].PVM_CPUload+HostInfo[i].NonPVM_CPUload)>=NODE_OVERLOAD_TRESHOLD){
 	      for(j=0;j<TaskInfo.size();j++){
		if(TaskInfo[j].Host==i){
		    for(k=0;k<HostInfo.size();k++){
	    	      if(HostInfo[k].Name!=""){
			if(!HostInfo[k].source){
			  /* do not look at the source node and also not at the overloaded nodes */
			  if((k!=i)&&((HostInfo[k].PVM_CPUload+HostInfo[k].NonPVM_CPUload)<NODE_OVERLOAD_TRESHOLD)){
			    /* score this particular solution */
			    score=ScoreMigration(TaskInfo,HostInfo,i,k,j);
			    cerr<<"src="<<HostInfo[i].Name<<" dest="<<HostInfo[k].Name<<" task_name="<<TaskInfo[j].task_name<<" score="<<score<<endl;
			    /* we have found a better score, so remember it */
			    if(score>best_score){
				best_score=score;
				best_src_node=i;
				best_dest_node=k;
				best_task=j;
			    }
			  }
			}
		      }
		    }
		}
	      }
	    }
	  }
        }
      }

      /* migrate only if we found something usefull */
      if((best_src_node<0)||(best_dest_node<0)||(best_task<0)||(best_score<=0))return found_mig;
      
      found_mig++;

      /* task to migrate */
#ifdef _DEBUG
      cerr<<"best_src_node="<<best_src_node<<" best_dest_node="<<best_dest_node<<" best_task="<<best_task<<endl;
      cerr<<"WON score="<<best_score<<" src="<<HostInfo[best_src_node].Name<<" dest="<<HostInfo[best_dest_node].Name<<" task="<<TaskInfo[best_task].PVMtid<<endl;
#endif    

      TMigration mg;
      mg.task_id = TaskInfo[best_task].PVMtid;
      mg.currentHostID = TaskInfo[best_task].Host;
      mg.newHostID = best_dest_node;
      mg.currentHostName = HostInfo[mg.currentHostID].Name;
      mg.newHostName = HostInfo[mg.newHostID].Name;

      mig_vec.push_back( mg );
      
      UpdateMigrationData(best_src_node,best_dest_node,best_task,TaskInfo,HostInfo);

      m--;
      
    }

    return found_mig;
}

#endif // SIMULATOR




/*=========================================================================*/
/* Input: Tasks - infos about the tasks currently running in the pool
 *        HostInfo - infos on the hosts available in the pool
 *        BestMapping - mapping to be achieved
 * Output: mig_vec - vector of all necessary migrations
 * Returns: mig_vec.size()
 */
static
unsigned
DetermineMigrations( const vector<TTaskInfo>& Tasks, const vector<THostInfo>& HostInfo,
                     const vector<unsigned>& BestMapping, vector<TMigration>& mig_vec )
{
    for( int i=0; i<BestMapping.size(); i++ ) {

        if( Tasks[i].Host == BestMapping[i] )   continue;

        TMigration m;
        m.task_id = Tasks[i].PVMtid;
        m.currentHostID = Tasks[i].Host;
        m.newHostID = BestMapping[i];
        m.currentHostName = HostInfo[m.currentHostID].Name;
        m.newHostName = HostInfo[m.newHostID].Name;
        mig_vec.push_back( m );
    }
    return mig_vec.size();
}


#ifndef SIMULATOR
/*=========================================================================*/
static
void
SetTimeExpired( int )
{
    TimeExpired = TRUE;
}


/*=========================================================================*/
/* Input: Secs - Number of seconds to wait before TimeExpired is set to TRUE
 */
static
void
SetAlarm( unsigned Secs )
{
    struct sigaction act;
    memset( &act, 0, sizeof(act) );
    act.sa_handler = SetTimeExpired;
    sigaction( SIGALRM, &act,0 );
    alarm( Secs );
}
#endif //SIMULATOR



/*=========================================================================*/
#ifdef DEBUG
static
void
ShowSolution( const string& Msg, vector<unsigned>& Mapping, unsigned Tasks,
              double cost, double bound )
{
    cout << Msg << " (" << cost << "," << bound << "):\t";
    for( int i=0; i<Tasks; i++ ) {
        cout << i << "->" << Mapping[i] << ",";
    }
    cout << endl;
}
#endif



//===================================================================
static
double
NewCosts( unsigned Task, unsigned NewHost,
            unsigned& MigrationsSoFar, double& MaxCpuSoFar, double& MaxMemSoFar,
            const vector<TTaskInfo>& TaskInfo,
            const vector<THostInfo>& HostInfo, vector<double>& PVM_CPUload,
            vector<double>& PVM_MEMload )
{
    assert( PVM_CPUload.size() == HostInfo.size() );
    assert( PVM_MEMload.size() == HostInfo.size() );

    // Costs caused by CPU load
    PVM_CPUload[NewHost] += TaskInfo[Task].CPUload / CpuFactor[NewHost];
    double CpuCostsOnHosts =
        (PVM_CPUload[NewHost]+HostInfo[NewHost].NonPVM_CPUload);

    // Costs caused by memory consumption
    PVM_MEMload[NewHost] += TaskInfo[Task].MEMload / MemFactor[NewHost];
    double MemCostsOnHosts(PVM_MEMload[NewHost]+HostInfo[NewHost].NonPVM_MEMload);

    // Contribution of this new task to the various load criterions
    MigrationsSoFar += (TaskInfo[Task].Host == NewHost) ? 0 : 1;
    MaxCpuSoFar = max( CpuCostsOnHosts, MaxCpuSoFar );
    MaxMemSoFar = max( MemCostsOnHosts, MaxMemSoFar );

    return MigrationsSoFar * MigWeight +
           MaxCpuSoFar * CpuWeight +
           MaxMemSoFar * MemWeight;
}



/*=========================================================================*/
/* Input: Tasks - infos about the tasks currently running in the pool
 *        Hosts - infos on the hosts available in the pool
 *        Bound - Costs of best valid solution we have found so far
 *        TasksMapped - Number of taskes mapped in this (partial) solution
 *        CurrentMapping - partial solution that has to be investigated further
 *        MigrationsSoFar - Number of migrations contained in the current partial solution
 *        MaxCpuSoFar - Maximum load on any of the hosts so far
 *        MaxMemSoFar - Maximum memory usage on any of the hosts so far
 *        PVM_CPUload - CPU load induced by the current partial solution
 *        PVM_MEMload - memory load induced by the current partial solution
 *        BestMapping - best valid solution already found
 *        Nodes - Number of times the cost function was called
 * Output: Bound, CostSoFar, BestMapping
 *        Nodes - Number of times the cost function was called
 */
static
void
BB( const vector<TTaskInfo>& Tasks, const vector<THostInfo>& Hosts,
    double& Bound, unsigned TasksMapped, vector<unsigned>& CurrentMapping,
    unsigned MigrationsSoFar, double MaxCpuSoFar, double MaxMemSoFar,
    const vector<double>& PVM_CPUload, const vector<double>& PVM_MEMload,
    vector<unsigned>& BestMapping, unsigned& Nodes )
{
#ifndef SIMULATOR
    extern t_global_options global_options;
#endif
    int i,j;
    double Cost = -1;

    assert( TasksMapped < Tasks.size() );

    for( i=Tasks[TasksMapped].Host,j=0; j<Hosts.size() && !TimeExpired; i=(i+1)%Hosts.size(),j++ ) {
        /* Try to map the next task onto host i */
        unsigned migrationsSoFar = MigrationsSoFar;
        double maxCpuSoFar = MaxCpuSoFar;
        double maxMemSoFar = MaxMemSoFar;
        vector<double> pvm_CPUload( PVM_CPUload );
        vector<double> pvm_MEMload( PVM_MEMload );
        CurrentMapping[ TasksMapped ] = i;
        ++Nodes;

        Cost = NewCosts( TasksMapped, i, migrationsSoFar, maxCpuSoFar, maxMemSoFar,
                                       Tasks, Hosts, pvm_CPUload, pvm_MEMload );


        /* If we are still cheaper than the bound, try further */
        if( Bound < 0  ||  Cost < Bound ) {
            if( TasksMapped+1 == Tasks.size() ) {
                /* We have a new best solution */
                BestMapping = CurrentMapping;
                Bound = Cost;
#ifdef DEBUG
                ShowSolution( "Full", CurrentMapping, TasksMapped+1, Cost, Bound );
#endif
#ifndef SIMULATOR
                // If this was the first solution, that the alarm timer
                SetAlarm( global_options.MaxSearchTime );
#endif
            } else {
                /* We still have to look deeper.. */
                BB( Tasks, Hosts, Bound, TasksMapped+1, CurrentMapping,
                    migrationsSoFar, maxCpuSoFar, maxMemSoFar,
                    pvm_CPUload, pvm_MEMload, BestMapping, Nodes );
            }
        } else {
            // We have reached a cut-off
#ifdef DEBUG
            ShowSolution( "Part", CurrentMapping, TasksMapped+1, Cost, Bound );
#endif
        }
    }
}


#ifdef USE_HASH_DATA
#include "balancer.h"
#endif

#ifndef SIMULATOR
/*=========================================================================*/
/* Input: task_stats  -  statistics about all pvm tasks
 * Output: TaskInfo - statiscal information about all moveable pvm tasks
 */
static
void
GetTaskInfo( procs_hosts_stats* task_stats, vector<TTaskInfo>& TaskInfo, unsigned hosts ) 
{
    int mypid = getpid();
    // For each host...
    for( int h=0; h<hosts; h++ ) {
        for( int t=0; t<task_stats[h].size; t++ ) {
            if( task_stats[h].base[t].task_is_dead ||
                !task_stats[h].base[t].stats.movable ||
                task_stats[h].base[t].stats.pid==mypid )    continue;
            // For each process on this host
            TTaskInfo task;
            task.Host = task_stats[h].base[t].stats.hid-1;
            task.PVMtid = task_stats[h].base[t].stats.tid;
            task.CPUload = max( 0.0 , task_stats[h].base[t].stats.pct_cpu * CpuFactor[h] );
            task.MEMload = max( 0.0, task_stats[h].base[t].stats.pct_mem * MemFactor[h] );
            task.DeadPointer = &task_stats[h].base[t].task_is_dead;

	    task.task_name=task_stats[h].base[t].stats.pname;
	    
	    task.lifespan=task_stats[h].base[t].stats.totals;
	    task.CPUload_avg=max( 0.0 , task_stats[h].base[t].stats.total_pct_cpu * CpuFactor[h] );
	    
#ifdef USE_HASH_DATA
	    struct bal_task_entry *itm;
	    
	    if((itm=bal_search(task.task_name))!=NULL){
		task.hash_load=itm->cpu_pct_avg;
		task.hash_life=itm->total_avg;
		task.isdata=1;
		
		if(task.hash_life-task.lifespan==0)task.load_expected=task.CPUload;
		  else{
		    task.load_expected=((task.hash_load/task.hash_life)-(task.CPUload_avg/task.lifespan))/(task.hash_life-task.lifespan);
		  }
	    }else{
	      task.isdata=0;
	      task.load_expected=task.CPUload;
	    }
#endif	    

            TaskInfo.push_back( task );
        }
    }
}
#endif // SIMULATOR




/*=========================================================================*/
static
void
HostInfoSubstractPVMLoad( vector<THostInfo>& HostInfo,
                          const vector<TTaskInfo>& TaskInfo )
{
    // Extract load of moveable PVM tasks
    for( int i=0; i<HostInfo.size(); i++ ) {
        if( HostInfo[i].Name == "" )       continue;

        // Determine load of all PVM processes on this host
        double PVM_CPUload = 0;
        double PVM_MEMload = 0;
        for( int t=0; t<TaskInfo.size(); t++ ) {
            if( TaskInfo[t].Host == i ) {
#ifdef USE_HASH_DATA
                PVM_CPUload += ((TaskInfo[t].CPUload+2*TaskInfo[t].load_expected)/3)/CpuFactor[i];
#else
                PVM_CPUload += TaskInfo[t].CPUload / CpuFactor[i];
#endif
                PVM_MEMload += TaskInfo[t].MEMload / MemFactor[i];
            }
        }

        // Substract PVM load from total load
        double h1 = HostInfo[i].NonPVM_CPUload;
        if( PVM_CPUload <= HostInfo[i].NonPVM_CPUload ) {
            HostInfo[i].NonPVM_CPUload -= PVM_CPUload;
        } else {
            HostInfo[i].NonPVM_CPUload = 0;
        }
	HostInfo[i].PVM_CPUload=PVM_CPUload;
	
        if( PVM_MEMload <= HostInfo[i].NonPVM_MEMload ) {
            HostInfo[i].NonPVM_MEMload -= PVM_MEMload;
        } else {
            HostInfo[i].NonPVM_MEMload = 0;
        }
	HostInfo[i].PVM_MEMload=PVM_MEMload;

        cout << "###1: " << HostInfo[i].Name << " (" << i << "): "
                << HostInfo[i].NonPVM_CPUload << " ("
                <<  PVM_CPUload << ", " << h1 << ")\n";
    }
}


#ifndef SIMULATOR
/*=========================================================================*/
/* Input: lm_stats    -  statistics about all available nodes
 *        hosts       -  number of available hosts
 * Output: HostInfo - statiscal information on load situation on all hosts
 */
static
void
GetHostInfo( load_stats** lm_stats, unsigned hosts, vector<THostInfo>& HostInfo,
             const vector<TTaskInfo>& TaskInfo )
{
    extern resdes* rd_base;

    assert( hosts == HostInfo.size() );

    // Parse the lm_stats structure
    for( int i=0; i<hosts; i++ ) {
        if( lm_stats[i] != 0 ) {
            int host = lm_stats[i]->stats.hid-1;
	    HostInfo[host].source=0;
	    HostInfo[host].dest=0;
            HostInfo[host].Name = lm_stats[i]->stats.hname;
            HostInfo[host].NonPVM_CPUload = lm_stats[i]->stats.cpuload;
            HostInfo[host].NonPVM_MEMload =
                ((double)lm_stats[i]->stats.uram + lm_stats[i]->stats.uswap ) /
                (lm_stats[i]->stats.uram+lm_stats[i]->stats.fram);
	    HostInfo[host].trend= lm_stats[i]->stats.trend;
        }
    }

    cerr << HostInfo.size() << " hosts found.\n";
    for( int i=0; i<HostInfo.size(); i++ ) {
      cerr << dec<<i << " (" << HostInfo[i].Name << ") CPU: " << HostInfo[i].NonPVM_CPUload
             << " MEM: " << HostInfo[i].NonPVM_MEMload
             << " CPU-Factor: " << CpuFactor[i] << " MEM-Factor: " << MemFactor[i] << endl;
    }

    // Find task id's of PVM demons
    int nhost, narch;
    struct pvmhostinfo* hostp;
    pvm_config( &nhost, &narch, &hostp );
    for( int i=0; i<nhost; i++ ) {
        int resid = get_resdes_hid( rd_base, hostp[i].hi_name );
        if( resid < 0 ) continue;
        HostInfo[resid-1].pvmd = hostp[i].hi_tid;
    }
}



/*=========================================================================*/
/* Input: pvmd - PVM task id of pvmd
 *        HostInfo - infos on the hosts available in the pool
 * Returns: host id as specified in the resource file
 */
static
int
PVMDID2RESID( unsigned pvmd, const vector<THostInfo>& HostInfo )
{
    for( int i=0; i<HostInfo.size(); i++ ) {
        if( HostInfo[i].pvmd == pvmd )  return i;
    }
    return -1;
}





/*=========================================================================*/
/* Input: Tasks - vector of all tasks the monitor thinks there are
 *        Hosts - infos on the hosts available in the pool
 * Output: Tasks - all the tasks both monitor and pvmd think there are
 * Note: The monitor's internal data are corrected by calling remove_task()
 */
static
void
EliminateDeadTasks( vector<TTaskInfo>& Tasks, const vector<THostInfo>& Hosts )
{
    extern int tid;                      /* Global PVM task id */
    extern int ConnectedToPVM;
    extern void remove_task( int, int );

    if( !ConnectedToPVM ) return;
    int ntask; // Number of tasks;
    struct pvmtaskinfo *taskaddr;  // Base pointer to pvmtaskinfo structure

    // Get information of local PVM tasks from PVMD
    if( pvm_tasks(0, &ntask, &taskaddr) < 0) return;
    // Sort the task vector to speed up the search operations
    vector<TTaskInfo> Alive;
    sort( Tasks.begin(), Tasks.end() );

    // Eliminate those tasks that are not confirmed by the pvmd
    for( int i=0; i<ntask; i++ ) {
        if( PVMDID2RESID( taskaddr[i].ti_host, Hosts ) < 0 ) continue;
        vector<TTaskInfo>::iterator found =
            lower_bound( Tasks.begin(), Tasks.end(),
                         TTaskInfo( taskaddr[i].ti_tid,
                                    PVMDID2RESID(taskaddr[i].ti_host, Hosts) ) );
        while( found != Tasks.end()  &&  (*found).PVMtid == taskaddr[i].ti_tid ) {
            if( (*found).Host == PVMDID2RESID(taskaddr[i].ti_host, Hosts) )  {
                Alive.push_back( *found );
                Tasks.erase( found );
                // Break the loop
                found = Tasks.end();
                /***hallo cerr << " no "; */
            } else {
                // The task is a zombie
                (*found).IsDead( true );
                ++found;
            }
        }
        /***hallo cerr << "problem." << endl; */
    }

    // Inform the rest of the monitor about its inconsistent data
    for( int i=0; i<Tasks.size(); i++ ) {
        remove_task( Tasks[i].PVMtid, Tasks[i].Host );
    }

    // Correct the tasks vector
    Tasks = Alive;
}
#endif // SIMULATOR


//==========================================================================
#ifdef SIMULATOR
static
void
ReadHosts( istream& in, vector<THostInfo>& HostInfo,
            vector<double>& CpuFactor, vector<double>& MemFactor )
{
    string dummy;
    int hosts;
    in >> hosts >> dummy >> dummy;
    for( int i=0; i<hosts; i++ ) {
        string hostname;
        double cpuload, memload, cpufactor, memfactor;
        in >> dummy >> hostname >> dummy >> cpuload >> dummy >> memload
           >> dummy >> cpufactor >> dummy >> memfactor;
        CpuFactor.push_back( cpufactor );
        MemFactor.push_back( memfactor );
        HostInfo.push_back( THostInfo( hostname, cpuload, memload ) );
    }
}


static
void
ReadWeights( istream& in,
             double& migweight, double& cpuweight, double& memweight )
{
    string dummy;
    in >> dummy >> dummy >> dummy >> migweight
       >> dummy >> dummy >> dummy >> dummy >> cpuweight
       >> dummy >> dummy >> dummy >> dummy >> memweight;
}



static
void
ReadTasks( istream& in, vector<TTaskInfo>& TaskInfo )
{
    string dummy;
    int tasks;
    in >> dummy >> dummy >> dummy >> dummy >> dummy >> dummy >> dummy;
    in >> tasks >> dummy >> dummy;
    for( int i=0; i<tasks; i++ ) {
        int host;
        unsigned tid;
        double cpuload, memload;
        in >> dummy >> dummy >> host >> dummy >> hex >> tid >> dec >> dummy >> cpuload
           >> dummy >> memload;
        TaskInfo.push_back( TTaskInfo( tid, host, cpuload, memload ) );
    }
}



int testit()
{
    vector<TMigration> mig_vec;

    unsigned MigrationsNecessary = 0;
    mig_vec = vector<TMigration>();

    ifstream input( "c:\\testit.txt" );
    if( !input.good() ) {
        cerr << "Could not open c:\\testit.txt." << endl;
        return -1;
    }
    vector<THostInfo> HostInfo;
    vector<TTaskInfo> TaskInfo;
    ReadHosts( input, HostInfo, CpuFactor, MemFactor );
    ReadWeights( input, MigWeight, CpuWeight, MemWeight );
    ReadTasks( input, TaskInfo );

    cerr << HostInfo.size() << " hosts found.\n";
    for( int i=0; i<HostInfo.size(); i++ ) {
      cerr << dec<<i << " (" << HostInfo[i].Name << ") CPU: " << HostInfo[i].NonPVM_CPUload
             << " MEM: " << HostInfo[i].NonPVM_MEMload
             << " CPU-Factor: " << CpuFactor[i] << " MEM-Factor: " << MemFactor[i] << endl;
    }

    HostInfoSubstractPVMLoad( HostInfo, TaskInfo );

    cerr << "Migration cost is " << dec<<MigWeight << endl
         << "CPU weight factor is " << CpuWeight << endl
         << "Memory weight factor is " << MemWeight << endl;
    cerr << TaskInfo.size() << " tasks found" << endl;
    for( int i=0; i<TaskInfo.size(); i++ ) {
      cerr << i << ") Host: " << TaskInfo[i].Host << " TID: " << hex<<TaskInfo[i].PVMtid
             << " CPU: " << TaskInfo[i].CPUload << " MEM: " << TaskInfo[i].MEMload
             << endl;
    }

    // Holds the partial solution during the branch and bound run
    vector<unsigned> CurrentMapping( TaskInfo.size(), 0 );

    // Best mapping found so far
    vector<unsigned> BestMapping;

    // Costs of the best valid solution that has so far been found
    double Bound = -1;

    /* Call branch&bound routine */
    vector<double> PVM_CPUload( HostInfo.size(), 0 );
    vector<double> PVM_MEMload( HostInfo.size(), 0 );
    unsigned Nodes = 0;
    BB( TaskInfo, HostInfo, Bound, 0, CurrentMapping, 0, 0, 0, PVM_CPUload, PVM_MEMload,
        BestMapping, Nodes );
    cerr << "Number of nodes checked: " << dec<<Nodes << endl;

    /* Determine the necessary migrations */
    MigrationsNecessary = DetermineMigrations( TaskInfo, HostInfo, BestMapping, mig_vec );

    cerr << "Planning to do " << mig_vec.size() << " migrations\n"<<flush;
    for( int l=0; l<mig_vec.size(); l++ ) {
      cerr << "  task " << hex << mig_vec[l].task_id << " from " << mig_vec[l].currentHostName
             << " to " << mig_vec[l].newHostName << " (" << mig_vec[l].newHostID
             << ") " << endl << flush;
    }

    getchar();
    return MigrationsNecessary;
}
#endif
