#ifndef _MIGRATOR_H
#define _MIGRATOR_H

using std::string;
using std::vector;

/* Structure that defines a migration */
typedef struct {
    unsigned long task_id;
    unsigned currentHostID;
    unsigned newHostID;
    string currentHostName;
    string newHostName;
} TMigration;


#include "monitor.h"

//=== Some local definitions ======
class TTaskInfo {
 public:
    /// PVM task id of this process
    unsigned PVMtid;

    /// ID of the host on which this task is currently being executed
    unsigned Host;

    /// CPU load of this task in percent of the "normal" machine
    double CPUload;

    /// Memory load of this task in percent of the "normal" machine
    double MEMload;
 
    /// pointer to the name of the task
    char *task_name;

    /// the lifespan  of this task
    double lifespan;
    
    /// the average CPU load over the whole task's lifespan, also normalized to "normal" machine ...
    double CPUload_avg;
    
    /// Pointer to the task_is_dead field of the internal data structures
    int* DeadPointer;
#ifdef USE_HASH_DATA

    /* denotes that there was actually found some data about this task in the HASH-table */
    int isdata;

    /* the average load of the task from the HASH-table ... */
    double hash_load;

    /* the average life of a task from the HASH-table ... */
    double hash_life;

    /* this is the expected load for this task according to the HASH-table data ... */
    double load_expected;

#endif
 public:
    TTaskInfo( unsigned tid=0, unsigned host=0, double cpu=0, double mem=0 ) :
        PVMtid(tid), Host(host), CPUload(cpu), MEMload(mem), DeadPointer(0) {}

    bool operator < ( const TTaskInfo& Other ) const 
        { return PVMtid < Other.PVMtid ||
                 PVMtid == Other.PVMtid && Host < Other.Host; }
 
    bool operator == ( const TTaskInfo& Other ) const
        { return PVMtid == Other.PVMtid  &&  Host == Other.Host; }

    bool IsDead( bool flag )
        { if( DeadPointer ) *DeadPointer = flag; return flag; }

};



class THostInfo {
 public:
    /// Name of this host
    string Name;

    /// Task id of the PVM demon
    unsigned pvmd;

    /// CPU load of all non-pvm processes on this host in percent of "normal" machine
    double NonPVM_CPUload;

    /// CPU load of all pvm processes on this host in percent of "normal" machine
    double PVM_CPUload;

    /// Memory load of non-pvm processes on this host in percent of "normal" machine
    double NonPVM_MEMload;

    /// Memory load of pvm processes on this host in percent of "normal" machine
    double PVM_MEMload;

    int index;

    /* set this flag, if this host was used as a source or destination for migration, if so, don't use it as a target 
       or source again ... */
    int source;
    int dest;

    double trend;

 public:
    THostInfo( const string& name="", double cpuload=0, double memload=0 ) :
        Name(name), NonPVM_CPUload(cpuload), NonPVM_MEMload( memload ), PVM_CPUload(0), PVM_MEMload(0) {}
};

    

#ifndef SIMULATOR
/*=========================================================================*/
/* Find the best mapping
 * Input: host_stats  -  statistics about all  PVM tasks
 *        lm_stats    -  statistics about all available nodes
 *        hosts       -  number of available hosts
 * Output: mig_vec    - vector of all necessary migrations
 * Returns: Number of necessary migrations
 *
 */
int
FindBestMapping( procs_hosts_stats* host_stats, load_stats** lm_stats, unsigned hosts,
                 vector<TMigration>& mig_vec );
#endif                 



/*=========================================================================*/
/* Input: Tasks - infos about the tasks currently running in the pool
 *        Hosts - infos on the hosts available in the pool
 *        Bound - Costs of best valid solution we have found so far
 *        TasksMapped - Number of taskes mapped in this (partial) solution
 *        CurrentMapping - partial solution that has to be investigated further
 *        MigrationsSoFar - Number of migrations contained in the current partial solution
 *        MaxCpuSoFar - Maximum load on any of the hosts so far
 *        MaxMemSoFar - Maximum memory usage on any of the hosts so far
 *        PVM_CPUload - CPU load induced by the current partial solution
 *        PVM_MEMload - memory load induced by the current partial solution
 *        BestMapping - best valid solution already found
 *        Nodes - Number of times the cost function was called
 * Output: Bound, CostSoFar, BestMapping
 *        Nodes - Number of times the cost function was called
 */
static
void
BB( const vector<TTaskInfo>& Tasks, const vector<THostInfo>& Hosts, 
    double& Bound, unsigned TasksMapped, vector<unsigned>& CurrentMapping, 
    unsigned MigrationsSoFar, double MaxCPUSoFar, double MaxMemSoFar,
    const vector<double>& PVM_CPUload, const vector<double>& PVM_MEMload,
    vector<unsigned>& BestMapping, unsigned& Nodes );




/*=========================================================================*/
/* Input: Tasks - infos about the tasks currently running in the pool
 *        HostInfo - infos on the hosts available in the pool
 *        BestMapping - mapping to be achieved
 * Output: mig_vec - vector of all necessary migrations
 * Returns: mig_vec.size()
 */
static
unsigned 
DetermineMigrations( const vector<TTaskInfo>& Tasks, const vector<THostInfo>& HostInfo,
                     const vector<unsigned>& BestMapping, vector<TMigration>& mig_vec );




/*=========================================================================*/
/* Input: Task - task that is to be placed
 *        NewHost - Host on which the task will be placed
 *        TaskInfo - infos about the tasks currently running in the pool
 *        HostInfo - infos on the hosts available in the pool
 * Returns: Costs caused by this particular placement
 */
static
double
CostOfTask( unsigned Task, unsigned NewHost, double OldCost, unsigned& MigrationsSoFar,
            const vector<TTaskInfo>& TaskInfo,
            const vector<THostInfo>& HostInfo, vector<double>& PVM_CPUload,
            vector<double>& PVM_MEMload );



#ifndef SIMULATOR
/*=========================================================================*/
/* Input: task_stats  -  statistics about all pvm tasks
 * Output: TaskInfo - statiscal information about all moveable pvm tasks
 */
static
void
GetTaskInfo( procs_hosts_stats* task_stats, vector<TTaskInfo>& TaskInfo, unsigned hosts ) ;



/*=========================================================================*/
/* Input: lm_stats    -  statistics about all available nodes
 *        hosts       -  number of available hosts
 * Output: HostInfo - statiscal information on load situation on all hosts
 */
static
void
GetHostInfo( load_stats** lm_stats, unsigned hosts, vector<THostInfo>& HostInfo,
             const vector<TTaskInfo>& TaskInfo );




/*=========================================================================*/
/* Input: Tasks - vector of all tasks the monitor thinks there are
 *        Hosts - infos on the hosts available in the pool
 * Output: Tasks - all the tasks both monitor and pvmd think there are
 * Note: The monitor's internal data are corrected by calling remove_task()
 */
static
void
EliminateDeadTasks( vector<TTaskInfo>& Tasks, const vector<THostInfo>& Hosts );




/*=========================================================================*/
static
void
HostInfoSubstractPVMLoad( vector<THostInfo>& HostInfo,
                          const vector<TTaskInfo>& TaskInfo );

int
FindMigrations( procs_hosts_stats* task_stats, load_stats** lm_stats, unsigned hosts,
                 vector<TMigration>& mig_vec  , int maxmig);
		 

#endif // SIMULATOR

#endif
