#ifndef MON_UTIL_H
#define MON_UTIL_H

#include <rpc/types.h>  /* for XDR */
#include <rpc/xdr.h>    /* for XDR */
#include <netdb.h>      /* for MAXHOSTNAMELEN */
#include <ctype.h>      /* for isspace */



/* Known message types */
typedef enum MTYPE {
  COMMM = 0,           /* Message from communication monitor */
  LOADM = 1,           /* Message from load monitor */
  PROCM = 2,           /* Message from process monitor */
  TERMS = 3,	       /* Message from the terminated process monitor */
  MAXM,                /* Max. number of possible message types +1 */
  UNKNOWN = 0xffffffff /* Unknown message */
} MTYPE;

/* Message identification structure, sent before each data structure */
typedef struct message_id {
  MTYPE mtype;         /* message type */
} message_id;

/* Encode/Decode load monitor message identification into XDR stream */
int xdr_mtype_id (XDR *xdrs, message_id *id);


/* Structure to hold resource description entry */
typedef struct resdes {
  int hid;                     /* The host id */
  char hname[MAXHOSTNAMELEN];  /* The hostname */
  char ip[16];                 /* The IP address */
  double cpu_weight;            /* CPU performance realitve to reference machine */
  double mem_weight;            /* memory performance realitve to reference machine */
  struct resdes *next;
} resdes;



/* Data structure for load monitors */
typedef struct lm_data {
  char *hname;         /* Base pointer to slave hostname */
  int hid;             /* Host id from resource file */
  long ncpu;           /* Number of CPUs */
  double cpuload;      /* Load of CPU's */
  double trend;	       /* trend of CPU load difference from last time ... */
  long uram;           /* Used physical ram in kbytes */
  long fram;           /* Free physical ram in kbytes */
  long uswap;          /* Used swap in kbytes */
  long fswap;          /* Free swap in kbytes */
  long ts;             /* Time stamp */
} lm_data;

/* Encode/Decode load monitor data into XDR stream */
int xdr_lm_data (XDR *xdrs, lm_data *data);


/* Data structure for process monitor */
typedef struct procs_data {
  char *hname;           /* Base pointer to pvm hostname */
  char pname[128];	 /* Name of the process */
  int hid;               /* Host id from resource file */
  int tid;               /* Global PVM task id */
  long pid;              /* Local process id */
  double pct_cpu;        /* Percentage of CPU load */
  double pct_mem;        /* Percentage of system memory used by process*/
  unsigned long mem_img; /* Size of process image in [kB] */
  unsigned long mem_res; /* Size of memory resources used by process in [kB] */
  long ts;               /* Time stamp */
  
  double totals;	 /* time that has been taken into account by calculating pct_cpu  or total_pct_cpu respectively */
			 /* is also the total lifetime of the process */
  double total_pct_cpu;	 /* the all time average of the cpu time from its start */
  
  int movable;           /* Process can be migrated */
} procs_data;

/* Encode/Decode process monitor data into XDR stream */
int xdr_procs_data (XDR *xdrs, procs_data *data);


/* Data structure for global options read from config file */
typedef struct {
  unsigned long MonPort;     /* Port on which master monitor listens */
  char MonHost[80];          /* Host on which master monitor resides */
  char StartSlaveCmd[256];   /* Executable of slave montior in NFS space */
  char DumpDir[256];         /* Directory into which dumps are written */
  double CPUweight;	     /* Weight factor the user put on CPU load */
  double MEMweight;          /* Weight factor the user put on memory load */
  double MIGweight;          /* Weight factor the user put on migration costs */
  unsigned MaxSearchTime;    /* Number of seconds allowed for looking for the best mapping */
} t_global_options;

/* Function for reading global options from config file */
int read_options( t_global_options* options, char* Filename );

/* Check, if pvm is running */
int pvm_is_running( void );

extern int get_resdes_hid (resdes *rd_base, char *hname);

typedef struct semaphore {
     pthread_mutex_t lock;
     pthread_cond_t  cond;
     int count;
} *sem_t;

struct mon_data {
	int mon_montype;
	int mon_msgtype;
	int mon_src;
	int mon_dst;
	int mon_len;
	struct timespec mon_stime;
	struct timespec mon_etime;
	struct mon_data *mon_next;
};

#define MON_TYPE_SEND	0
#define MON_TYPE_RECV	1


#define PVMDWAITTIME    10


struct mon_list {
	struct mon_data *head;
	struct mon_data *tail;
	sem_t avail;
	pthread_mutex_t lock;
	pthread_t thread;
	int flags;
};

#define MON_EXITING	0x0001
	

/* Mutex variable for synchronized send */
extern pthread_mutex_t send_mutex; 

/* Thread to monitor pvm process statistics */
extern pthread_t procs_thread;    

/* used for LINUX port -> go through a ascii-file in /proc */

char *skip_token(const char *p);

char *skip_ws(const char *p);
 
char *read_procname(const char *p,char *name);

extern size_t writeall( int fd, const void* buf, size_t nbyte );

#endif /* MON_UTIL_H */
