#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

#include <rpc/types.h>
#include <rpc/xdr.h>

#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/procfs.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <strings.h>

#include "pvm3.h"

#include "mon_global.h"
#include "monlib.h"


/* The socket where the data will be written to */
int sockfd;   

/* Mutex variable for synchronized send */
pthread_mutex_t send_mutex; 

/* Thread to monitor pvm process statistics */
pthread_t procs_thread;     

struct mon_list __pvm_ml;

sem_t
sem_create(int val)
{
	sem_t sem;

	sem = (sem_t)calloc(1, sizeof *sem);
	pthread_mutex_init(&sem->lock, NULL);
	pthread_cond_init(&sem->cond, NULL);
	sem->count = val;
	return sem;
}



int
sem_init(sem_t* sem, int val)
{
	*sem = sem_create(val);
	return 0;
}


void
sem_wait(sem_t* s)
{
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	if (sem->count > 0) {
		sem->count--;
		pthread_mutex_unlock(&sem->lock);
		return;
	}

	pthread_cond_wait(&sem->cond, &sem->lock);
	sem->count--;
	pthread_mutex_unlock(&sem->lock);
}


void
sem_signal(sem_t* s)
{
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	sem->count++;
	pthread_cond_signal(&sem->cond);
	pthread_mutex_unlock(&sem->lock);
}



int
sem_testwait(sem_t* s)
{
	int retval;
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	if (sem->count == 0) {
		pthread_mutex_unlock(&sem->lock);
		return 0;
	}
	sem->count--;
	pthread_mutex_unlock(&sem->lock);
	return 1;
}


int
sem_destroy(sem_t* s)
{
	sem_t sem = *s;

	pthread_cond_destroy(&sem->cond);
	pthread_mutex_destroy(&sem->lock);
	free(sem);
	return 0;
}



void
mon_add_data(struct mon_list* mp, int type, int code, int src, int dst, int len, 
             struct timespec* stime, struct timespec* etime)
{
	struct mon_data md;

	md.mon_montype = type;
	md.mon_msgtype = code;
	md.mon_src = src;
	md.mon_dst = dst;
	md.mon_len = len;
	md.mon_stime = *stime;
	md.mon_etime = *etime;
	monlist_addtail(mp, &md);
	sem_signal(&mp->avail);
}



int
mon_get_data(struct mon_list* mp, struct mon_data* mdp)
{
	struct mon_data md;

	sem_wait(&mp->avail);
	if (mp->head == NULL && mp->flags & MON_EXITING)
		return 0;
	monlist_remhead(mp, mdp);
	return 1;
}



struct mon_list *
monlist_create()
{
	struct mon_list *mp;

	mp = (struct mon_list *)malloc(sizeof (struct mon_list));
	monlist_init(mp);
	return mp;
}



void
monlist_destroy(struct mon_list* mp)
{
	sem_destroy(&mp->avail);
	pthread_mutex_destroy(&mp->lock);
}



void
monlist_init(struct mon_list* mp)
{
	mp->head = mp->tail = NULL;
	sem_init(&mp->avail, 0);
	pthread_mutex_init(&mp->lock, NULL);
	mp->flags = 0;
}




void
monlist_addtail(struct mon_list* mp, struct mon_data* dp)
{
	struct mon_data *tp;

	pthread_mutex_lock(&mp->lock);
	tp = (struct mon_data *)malloc(sizeof (struct mon_data));
	*tp = *dp;
	tp->mon_next = 0;
	if (mp->tail == NULL)
		mp->head = mp->tail = tp;
	else {
		mp->tail->mon_next = tp;
		mp->tail = tp;
	}
	pthread_mutex_unlock(&mp->lock);
}



void
monlist_exit(struct mon_list* mp)
{
	pthread_mutex_lock(&mp->lock);
	mp->flags |= MON_EXITING;
	pthread_mutex_unlock(&mp->lock);
	sem_signal(&mp->avail);
}


void
monlist_remhead(struct mon_list* mp, struct mon_data* dp)
{
	pthread_mutex_lock(&mp->lock);
	*dp = *mp->head;
	free(mp->head);
	if (mp->tail == mp->head)
		mp->tail = NULL;
	mp->head = dp->mon_next;
	pthread_mutex_unlock(&mp->lock);
}



int
xdr_timespec(struct XDR* xdrs, struct timespec* tsp)
{
	if (!xdr_long(xdrs, &tsp->tv_sec))
		return FALSE;
	if (!xdr_long(xdrs, &tsp->tv_nsec))
		return FALSE;
	return TRUE;
}



int
xdr_mondata(struct XDR* xdrs, struct mon_data*mdp)
{
	if (!xdr_int(xdrs, &mdp->mon_montype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_msgtype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_src))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_dst))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_len))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_stime))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_etime))
		return FALSE;
	return TRUE;
}



int mon_connect(char *host, int port)
{
  struct sockaddr_in sin;
  int i, s, e, family, one;
  struct hostent *hp, hent;
  struct in_addr saddr;
  char buffer[MAXHOSTNAMELEN+1];
  int herr;

  hp = gethostbyname_r(host, &hent, buffer, sizeof buffer, &herr);
  if (hp == 0) {
    (void)fprintf(stderr, "sync: unknown host: %s\n", host);
    return -1;
  }
  family = hp->h_addrtype;
  bcopy(hp->h_addr, (char *)&saddr, hp->h_length);
  
  for (i = 0; i < CONRETRCNT; i++) {
    
    one = 1;
    
    /*
     * Wonder why the socket must be recreated each time,
     * but without this it doesn't work on AIX.
     */
    if ((s = socket(family, SOCK_STREAM, 0)) < 0) {
      perror("client socket");
      return -1;
    }
    
    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*)&one, sizeof (int));
    sin.sin_addr = saddr;
    sin.sin_port = port;
#ifdef HAS_SINLEN
    sin.sin_len = sizeof (struct sockaddr_in);
#endif
    sin.sin_family = family;
    if ((e = connect(s, (struct sockaddr *)&sin, sizeof(sin))) < 0) {
      /* Error connecting to maser, wait a little time */
      close(s);
      fprintf (stderr, "Can't contact monitor master process ");
      fprintf (stderr, "retrying in %d seconds.\n", CONRETRDEL);
      sleep (CONRETRDEL);
    } else {
      /* We are connected */
      break;
    }
  }
  
  /* Really connected or retry count exceeded */
  if (e < 0) {
    perror("connect");
    close(s);
    return -1;
  }

  return s;
}


void *mon(void *v)
{
  struct mon_list *mp = (struct mon_list*)v;

  char *buf;
  XDR xdr;

  struct mon_data md;
  message_id mid;

  buf = (char*)malloc (sizeof (struct mon_data));

  /* Set the message type */
  mid.mtype = COMMM;

  while (mon_get_data (mp, &md)) {

    xdrmem_create (&xdr, buf, sizeof (struct mon_data), XDR_ENCODE);
    xdr_mondata (&xdr, &md);
    xdr_destroy (&xdr);

    /* Send the message */
    mon_sendmessage(&mid, (void *) buf, sizeof (struct mon_data));
  }
  
  /* No more data and MONEXITING flag set */
  monlist_destroy(mp);
}


int mon_createthread(struct mon_list *mlp)
{
  
  extern t_global_options global_options;

  /* Attributes for the process statistics thread */
  pthread_attr_t procs_attr;  

  pthread_attr_t attr;

  /* Initialize monitored pvm messages queue */
  monlist_init(mlp);
  
  /* Initialize thread objects with default values */
  pthread_attr_init(&attr);
  pthread_attr_init(&procs_attr);
  pthread_mutex_init (&send_mutex, NULL);
     
  /* Try to connect to the monitor master */
  if ((sockfd = mon_connect (global_options.MonHost, global_options.MonPort)) < 0) {
    return -1;
  }

  /* Create thread to monitor pvm messages */
  if (pthread_create (&mlp->thread, &attr, mon, mlp) < 0) {
    return -1;
  }

  /* Create thread to monitor local pvm process statistics */
  if (pthread_create (&procs_thread, &procs_attr, mon_procs, mlp) < 0) {
    return -1;
  }
  return 0;
}

void mon_terminatethread(struct mon_list *mlp)
{
  int status1, status2;
  
  /* Signal the threads to exit */
  monlist_exit(mlp);
	
  /* Wait for termination of the pvm message monitoring thread */
  if (pthread_join (mlp->thread, (void **) &status1) < 0) {
    bailout ("mon_terminatethread(), pthread_join() for mlp->thread failed");
  }
  
  /* Wait for termination of the process statistics thread */
  if (pthread_join (procs_thread, (void **) &status2) < 0) {
    bailout ("mon_terminatethread(), pthread_join() for procs_thread failed");
  }
  
  /* Close the socket to the monitor master */
  shutdown(sockfd, 1);
  close(sockfd);
}



void mon_sendmessage (message_id *mid, void *buffer, int bytes2write)
{
  int n;                    /* Bytes written by socket */

  XDR xdres_id;             /* XDR encode stream for message ident. */
  char *xdrbuff_id;         /* XDR buffer for message identification */
  int xdrbs_id;             /* XDR buffer size for message ident. */
  
  
  /* Aquire the mutex */
  pthread_mutex_lock (&send_mutex);
  
  /* Allocate XDR buffer for the message identification */
  xdrbs_id = sizeof (message_id);
  if ((xdrbuff_id = 
       (char *) calloc (xdrbs_id, sizeof (char))) == NULL) {
    bailout ("mon_sendmessage(), calloc() for xdrbuff_id failed");
  }
  
  /* Create XDR stream for message identification */
  xdrmem_create (&xdres_id, xdrbuff_id, xdrbs_id, XDR_ENCODE);
  
  /* Encode the message type identifier */
  if (xdr_mtype_id (&xdres_id, mid) != TRUE) {
    bailout ("mon_sendmessage(), xdr_mtype_id() failed");
  }
  xdr_destroy(&xdres_id);
  
  /* Send message identification */
  if ((n = write (sockfd, xdrbuff_id, xdrbs_id)) < 0) {
    bailout ("mon_sendmessage(), write() for xdrbuff_id failed");
  } else {
    if (n != xdrbs_id) {
      fprintf (stderr, "Wrote only %d bytes of %d bytes\n", n, xdrbs_id);
      bailout ("mon_sendmessage(), write() for xdrbuff_id failed");
    }
  }
  
  /* Send the buffer */
  if ((n = write (sockfd, buffer, bytes2write)) < 0) {
    bailout ("mon_sendmessage(), write() for buffer failed");
  } else {
    if (n != bytes2write) {
      fprintf (stderr, "Wrote only %d bytes of %d bytes\n", n, bytes2write);
      bailout ("mon_sendmessage(), write() for buffer failed");
    }
  }
  
  /* Free the mutex */
  pthread_mutex_unlock (&send_mutex);
}


/* Thread for monitoring local pvm process statistics */
void *mon_procs(void *p)
{
}
    



